'''
Run 2*n+1 simulations of multi-zone building at each step in forecast horizon
to determine the heating/cooling/electric load deviation from the nominal operation.
'''


from eagers.building.basic.leward_calc import leward_calc
from eagers.building.basic.wind_speed_calc import wind_speed_calc
from eagers.building.basic.thermostat_humidistat import thermostat_humidistat
from eagers.building.basic.area_weighted_temperature import area_weighted_temperature
from eagers.building.process.load_sched import load_sched
from eagers.building.process.initial_param import initial_param
from eagers.building.process.reset_loop_param import reset_loop_param
from eagers.building.process.update_model_states_implicit import update_model_states_implicit
from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.ideal_hvac import ideal_hvac
from eagers.building.process.sim_air_loops import sim_air_loops
from eagers.building.process.net_water_loop_loads import net_water_loop_loads
from eagers.building.process.lin_load_increase import lin_load_increase
from eagers.building.components.unitary_systems import unitary_systems
from eagers.building.components.air_terminals import air_terminals

import datetime

def forecast_dr_capacity(buildings,observer,weather,date):
    n_b = len(buildings)
    if n_b==0:
        forecast = None
        T_building_effective = None
        heating = None
        cooling = None
    else:
        n = 3 #resolution for linear interpolation
        order = [j for j in range(n,0,-1)]
        order.append(0)
        order.extend([j for j in range(n+1,2*n+1)])
        if len(date) == 1:
            date = [date[0] - datetime.timedelta(hours=1),date[0]]
        forecast = {}
        n_s = len(date)-1
        # m_param = ['schedules','T_set','w_set','loads','gains','occupancy','mixing','infiltration','T_mains','water_heat','S'] #parameters that could be overridden by 'actual' measurements
        # forecast['m_param'] = [[{} for k in range(n_b)] for t in range(n_s)]
        f_keys = ['E0','C0','H0','T_avg','Tzone','Discomfort','Tmin','Tmax','H2E','C2E','th_bar','tc_bar','ua_c','ua_h','h_min','c_min','internal_gain']
        for f in f_keys:
            forecast[f] = [[0 for b in range(n_b)] for t in range(n_s)]
        forecast['Cap'] = [[] for b in range(n_b)]
        forecast['deadband'] = [[False for b in range(n_b)] for t in range(n_s)]
        forecast['Tz_nominal'] = [[] for b in range(n_b)]
        heating = [[[0 for j in range(2*n+1)] for b in range(n_b)] for t in range(n_s)]
        cooling=  [[[0 for j in range(2*n+1)] for b in range(n_b)] for t in range(n_s)]
        air_density = 1.204 #kg/m^3 #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
        cat = {'heat_elec''heat_gas''cool_elec''fan_elec''water_gas''water_elec''tower_elec''pumps''other_fans'}
        default = {}
        default['Tzone'] = []
        T_building_effective = [[[0 for i in range(2*n+1)] for k in range(n_b)] for j in range(n_s)]
        hvac_elec = [0 for i in range(2*n+1)]
        Tzone_dr = [[0 for i in range(2*n+1)] for j in range(n_s)]
        Tsurf_dr = [[0 for i in range(2*n+1)] for j in range(n_s)]
        humidity_dr = [[0 for i in range(2*n+1)] for j in range(n_s)]
        flow_param = {} #output for debugging
        flow_param['zone_flow'] = [[0 for i in range(2*n+1)] for j in range(n_s)]
        flow_param['zone_temp'] = [[0 for i in range(2*n+1)] for j in range(n_s)]
        flow_param['zone_transfer'] = [[0 for i in range(2*n+1)] for j in range(n_s)]
        cat = ['heat_elec','heat_gas','cool_elec','fan_elec','other_fans','water_gas','water_elec','tower_elec','pumps']
        e_use={}
        weather_now = {}
        w = {}

    for k in range(n_b):
        building = buildings[k]
        nz = len(building.zones['name'])
        frost = [[0 for i in range(len(building.cases))]] #need to make as initial condition that is passed from previous time step
        
        Tz0 = [Tz for Tz in observer['building_zone_temp'][k]]
        Ts0 = [Ts for Ts in observer['building_surf_temp'][k]]
        w0 = [j for j in observer['building_humidity'][k]]
        T_supply = observer['building_supply'][k]  #hot and cold water half loop tank temperature

        air_flow = [[0 for i in range(len(building.zones['name']))] for j in range(2*n+1)]
        Tcentral_supply = [[0 for i in range(len(building.zones['name']))] for j in range(2*n+1)]

        electric = [[0 for i in range(2*n+1)] for i in range(n_s)]
        ## need a multiplier on capacitance to acccount for walls
        forecast['Cap'][k] = sum([air_density*(1006 + 1860*w0[z])*building.zones['volume'][z]*building.zones['multiplier'][z]/1000 for z in range(nz)])#kJ/K
        T,air_nodes,plant_nodes,_,_ = initial_param(building,Tz0,Ts0)
        for l in range(len(building.plant_loop['name'])):
            plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][l]] = T_supply[l]
        
        t_last = date[0]
        for t in range(n_s):
            dt = (date[t+1] - t_last).seconds
            if dt <0:
                dt = dt + 24*3600 #necessary for warm-up period when date jumps back to start of day
            for i in weather:
                if len(weather[i])== len(date):
                    weather_now[i] = weather[i][t+1]
                else:
                    weather_now[i] = weather[i][t]
            leward = leward_calc(weather_now['wdir'],building.surfaces['exterior']['normal'])
            wind_speed = wind_speed_calc(weather_now['wspd'],building.ctf['s_height'],building.site['terrain'])
            schedules = load_sched(building.schedule,building.holidays,[date[t+1]],None,False)
            
            ###Load values specific to this moment in time
            T_set,w_set = thermostat_humidistat(building,schedules,Tz0) ## HVAC temperature Schedules
            loads,gains,occupancy,mixing,infiltration,S,frost_t,T_mains,water_heat = zone_loads(building,date[t+1],schedules,weather_now,Tz0,w0,dt,frost[t])

            forecast['Tmin'][t][k] = area_weighted_temperature(building,T_set['heat'])-1
            forecast['Tmax'][t][k] = max(forecast['Tmin'][t][k]+1,area_weighted_temperature(building,T_set['cool']))+1
            
            Qheat = [[] for j in range(2*n+1)]
            Qcool = [[] for j in range(2*n+1)]
            for j in range(2*n+1):
                T['zone'] = [Tz for Tz in Tz0]
                T['surf'] = [Ts for Ts in Ts0]
                w['zone'] = [wz for wz in w0]
                T,w,air_nodes,plant_nodes = reset_loop_param(building,T,w,air_nodes,plant_nodes,weather_now,date[t+1])
                for x in cat:
                    e_use[x] = 0
                T_set_mod = modify_T_set(T_set,j,n,Tzone_dr[t][order[0]])

                ###Determine HVAC setpoints
                air_nodes,central_flow,direct_flow,mixed_air_loop,direct_return_air,plenum,T,w,fan_avail = ideal_hvac(building,schedules,leward,wind_speed,S,air_nodes,weather_now['t_dryb'],mixing,infiltration,gains,T,w,T_set_mod,w_set,occupancy,dt,False)
                #no iteration, because we are holding the supply temperature fixed during the forecast
                supply_flow,e_use,plant_nodes = sim_air_loops(building,air_nodes,mixed_air_loop,central_flow,w['loop'],weather_now['t_dryb'],w['air'],plant_nodes,e_use,fan_avail)
                supply_flow,e_use,plant_nodes = air_terminals(building,air_nodes,supply_flow,e_use,plant_nodes,T['central'])
                supply_flow,plant_nodes,e_use = unitary_systems(building,air_nodes,supply_flow,direct_flow,direct_return_air,T['direct'],weather_now['t_dryb'],w['air'],fan_avail,plant_nodes,e_use)#zone equipment not part of central air loops
                # Collect loads to create vertices
                Qheat[order[j]], Qcool[order[j]],e_use['pumps']  = net_water_loop_loads(building,schedules,plant_nodes,T_mains,T_supply,dt)
                e_use['water_gas'] += water_heat/building.impact_factor['steam_efficiency'][0] ##unconnected water heat is considered purchased heat
                hvac_elec[j] = e_use['pumps'] + e_use['heat_elec'] + e_use['cool_elec'] + e_use['fan_elec']#net_water_loop loads did not simulate chillers, so this is cool elec not from chillers

                nonHVAC_electric = sum([loads['lighting_internal'][z]*loads['multiplier'][z] for z in range(nz)]) + sum([loads['plug_load'][z]*loads['multiplier'][z] for z in range(nz)]) + loads['exterior']['lighting'] + loads['exterior']['equipment'] + sum(loads['case']['electric']) + sum(loads['rack']['electric'])
                electric[t][order[j]] = (nonHVAC_electric + e_use['pumps'] + e_use['heat_elec'] + e_use['cool_elec'] + e_use['fan_elec'] + e_use['water_elec'] + e_use['tower_elec'])
                ## Simulate step forward in time   
                Tzone_dr[t][order[j]],Tsurf_dr[t][order[j]],humidity_dr[t][order[j]],e_dict = update_model_states_implicit(building,T,S,wind_speed,leward,gains,dt,w,mixing,plenum,infiltration,supply_flow)              
                T_building_effective[t][k][order[j]] = area_weighted_temperature(building,Tzone_dr[t][order[j]])
                air_flow[j] = [supply_flow['flow'][z]*building.zones['multiplier'][z] for z in range(nz)] #useful to compare total treated air flows
                Tcentral_supply[j] = supply_flow['Tdb']
                flow_param['zone_flow'][t][order[j]] = supply_flow['m_dot']
                flow_param['zone_temp'][t][order[j]] = supply_flow['Tdb']
                n_z = len(supply_flow['Tdb'])
                cp_zone = [(1006 + 1860*supply_flow['w'][z])*air_density for z in range(n_z)] #Specific heat in J/m^3
                flow_param['zone_transfer'][t][order[j]] = [supply_flow['m_dot'][z]*cp_zone[z]*(supply_flow['Tdb'][z] - T_set_mod['cool'][z]) for z in range(n_z)]

            forecast['tc_bar'][t][k],forecast['ua_c'][t][k],forecast['c_min'][t][k],forecast['C2E'][t][k] = lin_load_increase(T_building_effective[t][k],[sum(Qcool[j])/1000 for j in range(2*n+1)],[hvac_elec[j]/1000 for j in range(2*n+1)])
            forecast['th_bar'][t][k],forecast['ua_h'][t][k],forecast['h_min'][t][k],forecast['H2E'][t][k] = lin_load_increase(T_building_effective[t][k],[sum(Qheat[j])/1000 for j in range(2*n+1)],[hvac_elec[j]/1000 for j in range(2*n+1)])
            flat_start = max([forecast['tc_bar'][t][k][0],forecast['th_bar'][t][k][0]])
            flat_end = min([forecast['tc_bar'][t][k][1],forecast['th_bar'][t][k][1]])
            if (flat_start>forecast['Tmin'][t][k] or flat_end<forecast['Tmax'][t][k]) and flat_start<flat_end:
                forecast['deadband'][t][k] = True

            Tz0 = [Tz for Tz in Tzone_dr[t][order[0]]]
            Ts0 = [Ts for Ts in Tsurf_dr[t][order[0]]]
            w0 = [j for j in humidity_dr[t][order[0]]]
            heating[t][k] = [sum(Qheat[j]) for j in range(2*n+1)]
            cooling[t][k] = [sum(Qcool[j]) for j in range(2*n+1)]

            forecast['Discomfort'][t][k] = sum([max([1,occupancy[z]]) * building.zones['multiplier'][z] for z in range(nz)]) # how many people are made uncomfortable
            forecast['E0'][t][k] = electric[t][order[0]]/1000 #kW
            forecast['C0'][t][k] = sum(Qcool[order[0]])/1000 #kW
            forecast['H0'][t][k] = sum(Qheat[order[0]])/1000 #kW
            forecast['Tzone'][t][k] = [tz for tz in Tzone_dr[t][order[0]]]
            forecast['T_avg'][t][k] = area_weighted_temperature(building,Tzone_dr[t][order[0]])
            #ensure when heating = H0, & cooling = C0 that there is no change in electrical demand
            forecast['E0'][t][k] = max([0,forecast['E0'][t][k]- forecast['C0'][t][k]*forecast['C2E'][t][k] - forecast['H0'][t][k]*forecast['H2E'][t][k]])
            frost.append(frost_t)
            t= t+1
            t_last = date[t]
        forecast['Tz_nominal'][k] = Tzone_dr[0][order[0]]
    return forecast,T_building_effective,heating,cooling


def modify_T_set(T_set,j,n,T0):
    T_set_mod = {}
    T_set_mod['no_hvac'] = [j for j in T_set['no_hvac']]
    T_set_mod['heat'] = [j for j in T_set['heat']]
    T_set_mod['cool'] = [j for j in T_set['cool']]
    T_set_mod['dual_sp'] = True
    if j>0:
        for z in range(len(T_set['cool'])):
            if not T_set['no_hvac'][z] :
                r = 2 #degrees it varies +,- from default
                if j<=n:
                    T_set_mod['cool'][z] = T0[z] - j/n*r
                else:
                    T_set_mod['cool'][z] = T0[z] + (j-n)/n*r
                T_set_mod['heat'][z] = T_set_mod['cool'][z]
        T_set_mod['dual_sp'] = False
    return T_set_mod


from eagers.building.components.water_coil import water_coil

def air_terminals(building,air_nodes,zone_air,e_use,plant_nodes,terminal_T_supply):
    ## re-heat or zone box cooling 
    air_density = 1.204 #kg/m^3 #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    for j in range(len(building.hvac['terminals']['name'])):
        if building.hvac['terminals']['type'][j]=='SingleDuct:Uncontrolled':
            pass###
        elif building.hvac['terminals']['type'][j]=='SingleDuct:ConstantVolume:Reheat':
            pass###
        elif building.hvac['terminals']['type'][j]=='SingleDuct:VAV:Reheat':
            z = building.hvac['terminals']['zone'][j]
            if terminal_T_supply[z]>zone_air['Tdb'][z]:
                air = {}
                for x in zone_air:
                    air[x] = zone_air[x][z]
                air['m_dot'] = air['m_dot']*building.zones['multiplier'][z]
                #update intermediate air as it passes between equipment
                name = building.hvac['terminals']['reheat_coil_name'][j]
                c = building.coils_heating[name]
                if c['type'] == 'Heating:Fuel' or c['type']=='Heating:Electric':
                        cp_air = (1006 + 1860*air['w'])*air_density #Specific heat in J/m^3
                        Q_terminal = air['m_dot']*cp_air*(terminal_T_supply[z] - air['Tdb'])
                        air['Tdb'] = terminal_T_supply[z]
                        if c['type'] == 'Heating:Fuel':
                            e_use['heat_gas'] += Q_terminal/c['efficiency'] #heat added in W
                        elif c['type']=='Heating:Electric':
                            e_use['heat_elec'] += Q_terminal/c['efficiency'] #heat added in W
                elif c['type'] == 'Heating:Water' or c['type'] == 'Cooling:Water':
                        air,m_w,actual_load,out = water_coil(building,name,air,terminal_T_supply[z],plant_nodes)
                        plant_nodes['demand_flow'][out] = m_w #kg/s
                        plant_nodes['load'][out] = actual_load #load in Watts (negative because air leaves hotter)
                air['m_dot'] = air['m_dot']/building.zones['multiplier'][z]
                for x in zone_air:
                    zone_air[x][z] = air[x]
    return zone_air,e_use,plant_nodes


import datetime as dt

from eagers.basic.get_data import get_data
from eagers.basic.get_data import interp_data

def hydro_forecast(forecast,test_data,date,subnet,observer,options):
    #TODO rework for new nodal structure of forecast and prev_data/now_data
    if 'hydro' in subnet:
        freq = max([1,round(options['horizon']/24,0)]) #period of repetition (1 = 1 day) This is how far back the forecasting methods are able to see. It is irrelevant if the forecast is perfect
        d0 = date[0] - dt.timedelta(hours = 24*freq+options['resolution'])
        prev_date = [d0 + dt.timedelta(hours = i*options['resolution']) for i in range(round(24*freq/options['resolution'],0)+1)]
        prev_data = get_data(test_data,prev_date,None)
        n_s = len(prev_data['hydro']['inflow'])
        n = len(prev_data['hydro']['inflow'][0]) # Number of nodes
        now_data = get_data(test_data,date[0]-dt.timedelta(hours = options['resolution']),None)
        if observer['hydro_wy_timestamp'] !=None:
            xi = 0
            while observer['timestamp'][xi+1]>0 and observer['timestamp'][xi+1]<prev_data['timestamp'][0]:
                xi+=1
            xf = xi + 0
            while observer['timestamp'][xf+1]>0 and observer['timestamp'][xf+1]<now_data['timestamp'][0]:
                xf+=1
            yi = 0
            while prev_data['timestamp'][yi]<observer['timestamp'][0]:
                yi+=1
            prev_data['hydro']['outflow'] = [[0 for i in range(n)] for t in range(n_s)]
            now_data['hydro']['outflow'] =  [[0 for i in range(n)] for t in range(n_s)]
            #TODO revise to new node based data structure
            for n in range(len(subnet['hydro']['line_number'])):
                select_data = interp_data(observer['timestamp'],observer,prev_data['timestamp'][yi:],['lineflows'],n,options['resolution']*3600)
                prev_data['hydro']['outflow'][yi:][n] = select_data['lineflows'][0]
                now_data['hydro']['outflow'][n] = observer['lineflows'][xf][subnet['hydro']['lineNumber'][n]]
            forecast['hydro']['inflow'] = forecast_hydro(date,forecast['hydro_sourcesink'],subnet,prev_data,now_data)
        else:
            ##Only reached when finding initial operating condition: assume inflow = outflow
            if not 'inflow' in forecast['hydro']:
                forecast['hydro']['inflow'] = [[of[t] for t in range(len(of))] for of in forecast['hydro']['outflow']]
    return forecast

def forecast_hydro(date,source_sink,subnet,prev_data,now_data):
    # forecast the source/sink + previously released upsteam water flow at each 
    # time step: 1 to nS. When t<T it needs the scheduled outflow of the
    # upstream dam to be on the constant side of the optimization.
    # does not give a forecast at t=0;
    n_s = len(date)
    t_last_disp = now_data['timestamp']
    dtq1 = (date[0] - t_last_disp).total_seconds()
    in_flow = [[ss[t] for t in range(len(ss))] for ss in source_sink]
    outflow = {}
    outflow['timestamp'] = [prev_data['timestamp'],now_data['timestamp']]
    outflow['flow'] = [prev_data['hydro']['outFlow'],now_data['hydro']['outflow']]
    for n in range(len(subnet['hydro']['nodes'])):
        k = subnet['hydro']['up_river'][n]
        for j in range(len(k)):
            t_seg = subnet['hydro']['linetime'][k[j]]
            for t in range(n_s):
                if (date[t] - t_last_disp)<= t_seg/24: # All of the flow reaching this reservior at this time was previously scheduled
                    select_data = interp_data(outflow['timestamp'],outflow,date[t]-t_seg/24,['flow'],n,dtq1)
                    in_flow[t][n] = in_flow[t][n] + select_data['flow'][0] #outflow from previous dispatches because the time preceeds t_last_disp
                elif t>0 and (date[t-1] - t_last_disp)< t_seg/24: # A portion of the flow reaching this reservior at this time was previously scheduled
                    frac = (date[t]- t_last_disp - t_seg/24)/(date[t] - date[t-1])
                    if (date[t]-t_seg/24)>outflow['timestamp'][-1]:
                        in_flow[t][n] += (1-frac)*outflow[-1][k[j]]
                    else:
                        select_data = interp_data(outflow['timestamp'],outflow,date[t]-t_seg/24,['flow'],n,dtq1)
                        in_flow[t][n] += (1-frac)*select_data['flow'][0] #outflow from previous dispatches because the time preceeds t_last_disp
    return in_flow


"""Real-time dispatch plotting logic.

Functions:
new_figure
plot_component_dispatch
y_axis_scaling
assign_colors
convert_to_steps
"""

import math
import datetime as dt

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from eagers.config.plots import (
    DEFAULT_FIG_WIDTH,
    DEFAULT_FIG_HEIGHT,
    SUPTITLE_SPACE,
    cm_parula,
)
from eagers.basic.gen_limit import chp_heat


def new_figure(project_name, gen, subnet):
    n_nets = len(subnet['network_names'])
    fig = plt.figure(figsize=(DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT))
    axs = []
    gs = gridspec.GridSpec(n_nets, 1)
    fig.suptitle(f"Dispatch: {project_name}", fontsize='x-large')
    for i, net in enumerate(subnet['network_names']):
        # Create new Axes object.
        new_ax = fig.add_subplot(gs[i])
        new_ax.set_title(net.title().replace('_', ' '))
        has_stor = False
        for j in range(len(gen)):
            if 'stor' in gen[j] and net in gen[j]['subnet_node']:
                has_stor = True
                break
        if has_stor:
            # Instantiate a second axes that shares the same x-axis.
            axs.append([new_ax, new_ax.twinx()])
        else:
            axs.append([new_ax])

    # Constrain layout properly.
    gs.tight_layout(fig, rect=[0, 0, 1, 1 - SUPTITLE_SPACE])
    # Remove warning about tight_layout.
    fig.set_tight_layout(False)

    plt.ion()
    plt.show()
    return fig, axs


def plot_component_dispatch(axs, history, future, gen, subnet):
    date = [j for j in history['timestamp']]
    date.extend(future['timestamp'])
    color = assign_colors(len(gen), cm_parula)
    Ymax = []
    Ymin = []
    Yticks = []
    for j, net in enumerate(subnet['network_names']):
        gen_state = {}
        stor_state = {}
        c_plot = []
        ax1 = axs[j][0]
        for i, g in enumerate(gen):
            if any(g['name'] in k for k in subnet[net]['equipment']):
                # Get history and future data
                k = g['name']
                gen_state[k] = [j for j in history['generator_state'][k]]
                gen_state[k].extend([j for j in future['generator_state'][k]])
                if g['type'] == 'ACDCConverter':
                    if net == 'direct_current':
                        gen_state[k] = [j*g['output']['e'][0][1] if j>0 else j for j in gen_state[k]]
                    elif net == 'electrical':
                        gen_state[k] = [-j*g['output']['dc'][0][0] if j<0 else -j for j in gen_state[k]]
                if g['type'] == 'CombinedHeatPower' and net == 'district_heat':
                    gen_state[k] = chp_heat(g,gen_state[k])
                if 'stor' in g:
                    unusable = g['stor']['size'] - g['stor']['usable_size']
                    stor_state[k] = [j + unusable for j in history['storage_state'][k]]
                    stor_state[k].extend([j + unusable for j in future['storage_state'][k]])
                c_plot.append([j for j in color[i]])
        try:
            plt.sca(ax1)
        except ValueError:
            # Turn interactive mode off; it's no longer being used.
            plt.ioff()
        plt.cla()
        gen_state, units, yma, ymi, yti = y_axis_scaling(gen_state, net, date)
        Ymax.append(yma)
        Ymin.append(ymi)
        Yticks.append(yti)
        if yti[0] < 0:
            g_names = list(gen_state.keys())
            for k in range(len(c_plot)):
                if any(
                    [
                        sum([gen_state[g_names[f]][t] for f in range(k - 1)])
                        < 0
                        for t in range(len(date))
                    ]
                ):
                    c_plot[k].append(
                        0.5
                    )  # appending alpha to show overlaping negative values
        ax1.set_title(net.title().replace('_', ' '))
        ax1.set_ylabel(net + ' (' + units[0] + ')')
        date_step, gen_state_step = convert_to_steps(date, gen_state)
        ax1.stackplot(
            date_step,
            gen_state_step.values(),
            labels=gen_state.keys(),
            colors=c_plot,
        )
        ax1.set_xlim(date[0], date[-1])
        ax1.set_ylim(bottom=Ymin[j], top=Ymax[j])
        ax1.set_yticks(ticks=Yticks[j])
        stor_names = list(stor_state.keys())
        if len(axs[j]) > 1:
            ax2 = axs[j][1]
            try:
                plt.sca(ax2)
            except ValueError:
                # Turn interactive mode off; it's no longer being used.
                plt.ioff()
            plt.cla()
            stor_state, units2, yma, ymi, yti = y_axis_scaling(
                stor_state, net, date
            )
        # date_step,stor_state_step = convert_to_steps(date,stor_state)
        for sn in stor_names:
            ax2.set_ylabel(
                'State-of-Charge' + units2[1], color='tab:blue'
            )  # we already handled the x-label with ax1
            # ax2.plot(date_step,stor_state_step[sn],label= sn+'_SOC', color='tab:blue')
            ax2.plot(date, stor_state[sn], label=sn + '_SOC', color='tab:blue')
            ax2.tick_params(axis='y', labelcolor='tab:blue')
            ax2.set_ylim(ymi, yma)
            ax2.set_yticks(ticks=yti)
        ax1.legend(loc='upper right')
        ax1.plot(
            [history['timestamp'][-1], history['timestamp'][-1]],
            [-1e8, 1e8],
            color='tab:cyan',
        )

    plt.draw()
    plt.pause(0.001)


def y_axis_scaling(data, net, date):
    val = [
        sum(data[f][t] for f in list(data.keys()))
        for t in range(1, len(date))
    ]
    pos = max(val)
    neg = min(val)
    if pos > neg:
        oom = max(1, 0.2 + math.log10(pos - min(neg, 0)))
    else:
        oom = 1
    if oom > 6.30103:
        if net == 'hydro':
            units = ['1e9 cfs', 'Billion acre-ft']
        else:
            units = ['GW', 'GWh']
        oom = oom - 6
        for f in list(data.keys()):
            data[f] = [k / 1e6 for k in data[f]]
    elif oom > 3.30103:
        if net == 'hydro':
            units = ['1e6 cfs', 'Million acre-ft']
        else:
            units = ['MW', 'MWh']
        oom = oom - 3
        for f in list(data.keys()):
            data[f] = [k / 1e3 for k in data[f]]
    else:
        if net == 'hydro':
            units = ['1000 cfs', '1000 acre-ft']
        elif net == 'cold_water' or net == 'building_temperature':
            units = ['F', 'kWh']
        else:
            units = ['kW', 'kWh']
    if net == 'building_temperature':
        Ymax = 78
        Ymin = 56
        Yspace = 2
    else:
        if (
            oom - int(oom)
        ) == 0:  # count in increments of 1, 10, 100 or 1000 etc
            Yspace = 2.5 * 10 ** (oom - 1)
            Ymax = 10 ** oom
        elif (oom - int(oom)) > 0.6990:
            Yspace = 2.5 * 10 ** int(oom)
            Ymax = 10 ** int(1 + oom)
        elif (
            oom - int(oom)
        ) > 0.30103:  # count in increments of 5, 50, 500 or 5000 etc
            Yspace = 10 ** int(oom)
            Ymax = 0.5 * 10 ** int(1 + oom)
        else:  # count in increments of 2, 20, 200 or 2000 etc
            Yspace = 0.5 * 10 ** int(oom)
            Ymax = 0.2 * 10 ** int(1 + oom)

        negTicks = int(min(min(data[f]) for f in list(data.keys())) / Yspace)

        if negTicks >= 0:
            Ymin = 0
        else:
            Ymin = Yspace * negTicks
        nticks = int((Ymax - Ymin) / Yspace) + 1
        while nticks > 14:
            Yspace = 2 * Yspace
            nticks = int((Ymax - Ymin) / Yspace) + 1
            negTicks = int(
                min(min(data[f]) for f in list(data.keys())) / Yspace
            )
            Ymin = Yspace * negTicks

        Yticks = [Ymin + j * Yspace for j in range(nticks)]
    return data, units, Ymax, Ymin, Yticks


def assign_colors(n, cm_parula):
    n2 = len(cm_parula)
    color = []
    for i in range(n):
        v = i / n * n2
        r = v - int(v)
        color.append([
            (1 - r) * cm_parula[int(v)][j] + r * cm_parula[int(v + 0.5)][j]
            for j in range(3)
        ])
    return color


def convert_to_steps(date, original):
    names = list(original.keys())
    new = {}
    date_step = []
    for n in names:
        new[n] = []
        for t in range(len(date)):
            new[n].extend([original[n][t], original[n][t]])
    for t in range(len(date) - 1):
        date_step.extend([date[t], date[t + 1] - dt.timedelta(seconds=0.001)])

    date_step.append(date[-1])
    date_step.append(date[-1] + (date[-1] - date[-2]))
    return date_step, new



import numpy as np
from numpy import degrees, radians, sin, cos, tan, arcsin, arccos


def solar_calc(longitude, latitude, time_zone, inp_date):
    """ Calculate position of sun and sunrise and sunset times.\n
    Calculated using NOAA solar calculations available at:
    https://www.esrl.noaa.gov/gmd/grad/solcalc/NOAA_Solar_Calculations_day.xls \n
    LONGITUDE is longitude (+ to east).\n
    LATITUDE is latitude (+ to north).\n
    TIME_ZONE is the time zone (+ to east).\n
    INP_DATE is the date as a list of datetime objects, i.e. Jan 1 2017 = 
    datetime(2017,1,1).\n
    SUNRISE and SUNSET are given in fraction of the day, i.e. 6am = 6/24.\n
    AZIMUTH and ZENITH are given in degrees."""

    tpm = np.array([(x - x.replace(hour=0, minute=0, second=0)).seconds / 24 / 3600 \
        for x in inp_date]) # time past midnight [days]
    jd = [x.toordinal() + 1721424.5 + y - time_zone/24 for x, y in zip(inp_date, tpm)]
        # Julian day [days]
        # Julian calendar epoch: Jan 1, 4713 B.C., 12:00:00.0
        # UNIX epoch: Jan 1, 0001 A.D., 00:00:00.0
        # MATLAB epoch: Jan 1, 0000 A.D., 00:00:00.0
        # 1721423.5 days from Julian epoch to UNIX epoch, plus 1 to match
        #     Microsoft Excel's year 1900 exception to the Gregorian calendar.
        # USNO Julian Date Converter:
        #     http://aa.usno.navy.mil/data/docs/JulianDate.php
        # More info on Excel dates:
        #     http://calendars.wikia.com/wiki/Microsoft_Excel_day_number
    jc = (np.array(jd) - 2451545) / 36525 # Julian century
    geom_mean_long_sun = (280.46646 + jc * (36000.76983 + jc*0.0003032)) % 360
        # [degrees]
    geom_mean_anom_sun = 357.52911 + jc * (35999.05029 - 0.0001537*jc) # [degrees]
    eccent_earth_orbit = 0.016708634 - jc * (0.000042037 + 0.0000001267*jc)
    sun_eq_of_center = \
        sin(radians(geom_mean_anom_sun)) * (1.914602 - jc * (0.004817 + 0.000014*jc)) \
        + sin(radians(2*geom_mean_anom_sun)) * (0.019993 - 0.000101*jc) \
        + sin(radians(3*geom_mean_anom_sun)) * 0.000289
    sun_true_long = geom_mean_long_sun + sun_eq_of_center # [degrees]
    mean_obliq_ecliptic = 23 + (26 + (21.448 - jc * (46.815 \
        + jc * (0.00059 - jc*0.001813))) / 60) / 60 # [degrees]
    sun_app_long = sun_true_long - 0.00569 \
        - 0.00478 * sin(radians(125.04 - 1934.136*jc)) # [degrees]
    obliq_corr = mean_obliq_ecliptic \
        + 0.00256*cos(radians(125.04 - 1934.136*jc)) # [degrees]
    sun_declin = degrees(arcsin(sin(radians(obliq_corr)) \
        * sin(radians(sun_app_long)))) # solar declination [degrees]
    var_y = tan(radians(obliq_corr/2)) * tan(radians(obliq_corr/2))
    eq_of_time = 4 * degrees( \
        var_y * sin(2*radians(geom_mean_long_sun)) \
        - 2 * eccent_earth_orbit * sin(radians(geom_mean_anom_sun)) \
        + 4 * eccent_earth_orbit * var_y * sin(radians(geom_mean_anom_sun)) * cos(2*radians(geom_mean_long_sun)) \
        - 0.5 * var_y * var_y * sin(4*radians(geom_mean_long_sun)) \
        - 1.25 * eccent_earth_orbit * eccent_earth_orbit * sin(2*radians(geom_mean_anom_sun)))
    ha_sunrise = degrees(arccos( \
        cos(radians(90.833)) / (cos(radians(latitude)) * cos(radians(sun_declin))) \
        - tan(radians(latitude)) * tan(radians(sun_declin))))
        # sunlight hours [degrees]
    solar_noon = (720 - 4*longitude - eq_of_time + time_zone*60) / 1440
    sunrise = solar_noon - ha_sunrise * 4 / 1440 # Local Sidereal Time (LST)
    sunset = solar_noon + ha_sunrise * 4 / 1440 # Local Sidereal Time (LST)
    tst = (tpm*1440 + eq_of_time + 4*longitude - 60*time_zone) % 1440 # True Solar Time [min]
    hour_angle = [x/4 + 180 if x/4 < 0 else x/4 - 180 for x in tst] # [degrees]
    zenith = degrees(arccos( \
        sin(radians(latitude)) * sin(radians(sun_declin)) \
        + cos(radians(latitude)) * cos(radians(sun_declin)) * cos(radians(hour_angle))))  # [degrees]
    ang = degrees(arccos( \
        ((sin(radians(latitude))*cos(radians(zenith))) - sin(radians(sun_declin))) \
        / (cos(radians(latitude))*sin(radians(zenith)))))
    azimuth = np.array([(x+180) % 360 if y > 0 else (540-x) % 360 \
        for x, y in zip(ang, hour_angle)])
        # [degrees] (clockwise from N)
    return sunrise, sunset, azimuth, zenith



"""Logic for recording the result of one step of the dispatch loop.

Functions:
dispatch_record
record_initial_conditions
"""
from eagers.basic.net_cost import net_cost
from eagers.write.result_file import \
    append_step_dispatch, append_dispatch_ic, append_step_predicted


def dispatch_record(gen, dispatch, observer, market, subnet, actual_data, solution, project_name):
    """Records the last dispatch 
    """
    dispatch['timestamp'] = solution['timestamp'][0]
    for i,k in enumerate(observer['gen_names']):
        dispatch['generator_state'][k] = solution['generator_state'][k][0]
        if k in solution['storage_state']:
            dispatch['storage_state'][k] = solution['storage_state'][k][0]
    for i,k in enumerate(observer['line_names']):
        dispatch['line_flow'][k] = solution['line_flow'][k][0]
    for i in range(len(observer['building_avg_temp'])):
        k = observer['building_name'][i]
        dispatch['building'][k]['cooling'] = solution['building'][k]['cooling'][0]
        dispatch['building'][k]['heating'] = solution['building'][k]['heating'][0]
        dispatch['building'][k]['Tzone'] = observer['building_zone_temp'][i]
        dispatch['building'][k]['supply'] = observer['building_supply'][i]
        dispatch['building'][k]['return_'] = observer['building_return'][i]
        dispatch['building'][k]['avg_T'] = observer['building_avg_temp'][i]
        dispatch['building'][k]['electric'] = actual_data['building']['E0'][0][i]
        
    for i in range(len(observer['fluid_loop_name'])):
        dispatch['fluid_loop'][observer['fluid_loop_name'][i]] = observer['fluid_loop_temperature'][i]
    
    for net in subnet['network_names']:
        for an in subnet[net]['nodes']:
            for n in an:
                if n in actual_data and 'demand' in actual_data[n]:
                    dispatch['nodedata'][n]['demand'] = actual_data[n]['demand'][0]
    for k in list(actual_data['weather'].keys()):
        dispatch['weather'][k] = actual_data['weather'][k][0]
    if 'hydro_soc' in solution and len(solution['hydro_soc']) > 0:
        dispatch['out_flow'] = [solution['line_flow'][subnet['hydro']['line_number'][n]][0] for n in range(len(subnet['hydro']['nodes']))]
    # Write result of simulation step to HDF5 file.
    append_step_dispatch(project_name, dispatch)

def predict_record(gen, predicted, observer, market, subnet, date, forecast, solution, project_name):
    n_s = len(date) - 1  # Numer of steps.
    predicted['timestamp'] = forecast['timestamp']
    predicted['cost'] = net_cost(gen, solution['dispatch'], date, market)
    for i,k in enumerate(observer['gen_names']):
        predicted['generator_state'][k] = solution['generator_state'][k]
        if k in solution['storage_state']:
            predicted['storage_state'][k] = solution['storage_state'][k]
    for i in range(len(observer['building_avg_temp'])):
        k = observer['building_name'][i]
        predicted['building'][k]['Tzone'] = [forecast['building']['Tzone'][t][i] for t in range(n_s)]
        predicted['building'][k]['electric'] = [forecast['building']['E0'][t][i] for t in range(n_s)]
        predicted['building'][k]['cooling'] = [forecast['building']['C0'][t][i] for t in range(n_s)]
        predicted['building'][k]['heating'] = [forecast['building']['H0'][t][i] for t in range(n_s)]
        predicted['building'][k]['avg_T'] = [forecast['building']['T_avg'][t][i] for t in range(n_s)]
    for i,k in enumerate(observer['line_names']):
        predicted['line_flow'][k] = solution['line_flow'][k]
    for i,k in enumerate(observer['fluid_loop_name']):
        predicted['fluid_loop'][k] = solution['fluid_loop'][k]
    for net in subnet['network_names']:
            for an in subnet[net]['nodes']:
                for n in an:
                    if n in forecast and 'demand' in forecast[n]:
                        predicted['nodedata'][n]['demand'] = forecast[n]['demand']
    for k in list(forecast['weather'].keys()):
        predicted['weather'][k] = forecast['weather'][k]
    if 'lb_relax' in solution:
        predicted['lb_relax'] = solution['lb_relax']

    # Write result of simulation step to HDF5 file.
    append_step_predicted(project_name, predicted)


def record_initial_conditions(data_t0, dispatch, subnet, observer, project_name):
    dispatch['timestamp'] = data_t0['timestamp'][0]
    observer['history']['timestamp'] = [data_t0['timestamp'][0]]
    for i,k in enumerate(observer['history']['generator_state']):
        dispatch['generator_state'][k] = observer['gen_state'][i]
        observer['history']['generator_state'][k] = [observer['gen_state'][i]]
        if k in observer['history']['storage_state']:
            dispatch['storage_state'][k] = observer['stor_state'][i]
            observer['history']['storage_state'][k] = [observer['stor_state'][i]]
    if len(observer['building_name'])>0:
        for i in range(len(observer['building_name'])):
            k = observer['building_name'][i]
            A = observer['building_conditioned_floor_area'][i]
            # Area-weighted temperature: Current average building zone
            # temperature (used as IC in optimization).
            t = sum([observer['building_zone_temp'][i][z]*A[z] for z in range(len(A))]) / sum(A)
            dispatch['building'][k]['avg_T'] = [t]
            observer['history']['building'][k]['avg_T'] = [t]
            dispatch['building'][k]['electric'] = [data_t0['building']['E0'][0][i]]
            dispatch['building'][k]['cooling'] = [data_t0['building']['C0'][0][i]]
            dispatch['building'][k]['heating'] = [data_t0['building']['H0'][0][i]]
            dispatch['building'][k]['Tzone'] = [observer['building_zone_temp'][i]]
            dispatch['building'][k]['return_'] = [observer['building_return'][i]]
            dispatch['building'][k]['supply'] = [observer['building_supply'][i]]
    for net in subnet['network_names']:
        for an in subnet[net]['nodes']:
            for n in an:
                if n in data_t0:
                    headers = list(data_t0[n].keys())
                    headers.remove('network')
                    for h in headers:
                        dispatch['nodedata'][n][h] = data_t0[n][h][0]
                        observer['history']['nodedata'][n][h] = [data_t0[n][h][0]]
    for k in data_t0['weather'].keys():
        dispatch['weather'][k] = data_t0['weather'][k][0]
        observer['history']['weather'][k] = [data_t0['weather'][k][0]]
    if not project_name is None:
        # Write to file.
        append_dispatch_ic(project_name, dispatch)


from eagers.building.components.cooling_coil import cooling_coil
from eagers.building.components.fan_calc import fan_calc
from eagers.building.components.water_coil import water_coil
from eagers.building.basic.psychometric import psychometric

def unitary_heat_cool(building,fan_avail,air,T_supply,T_air,e_use,plant_nodes,uhc):
    name = uhc['cool_coil']
    if air['m_dot']>1e-8:
        if 'cool_coil_type' in uhc:
            if (uhc['cool_coil_type']=='Coil:Cooling:DX:TwoSpeed' or
                uhc['cool_coil_type']=='Coil:Cooling:DX:SingleSpeed'):
                    air,P_coil = cooling_coil(building,name,air,T_air,T_supply)
                    e_use['cool_elec'] += P_coil
            elif (uhc['cool_coil_type']=='Coil:Cooling:Water:DetailedGeometry' or
                  uhc['cool_coil_type']=='Coil:Cooling:Water'):
                    air,m_w,actual_load,out = water_coil(building,name,air,T_supply,plant_nodes)
                    plant_nodes.demand_flow[out] = m_w #kg/s
                    plant_nodes.load[out] = actual_load #load in Watts (negative because air leaves hotter)
        name = uhc['heat_coil']
        c = building.coils_heating['name'].index(uhc['heat_coil'])
        if building.coils_heating['type'][c]=='Fuel' or building.coils_heating['type'][c]=='Electric':
            cp_supply = 1006 + 1860*air['w']
            load_heat = max(0,air['m_dot']*cp_supply*(T_supply-air['Tdb']))
            if load_heat>0:
                air['Tdb'] += load_heat/(cp_supply*air['m_dot'])
                air['h'] = psychometric(air,'h')
                if building.coils_heating['type'][c]=='Fuel':
                    e_use['heat_gas'] += load_heat/building.coils_heating['efficiency'][c] #heat added in W
                else:
                    e_use['heat_elec'] += load_heat/building.coils_heating['efficiency'][c] #heat added in W
        elif building.coils_heating['type'][c]=='Water':
                air,m_w,actual_load,out = water_coil(building,uhc['heat_coil'],air,T_supply,plant_nodes)
                plant_nodes.demand_flow[out] = m_w #kg/s
                plant_nodes.load[out] = actual_load #load in Watts (negative because air leaves hotter)
    air,P_fan = fan_calc(building.fans[uhc['fan_name']],air,fan_avail[uhc['fan_name']])
    e_use['fan_elec'] += P_fan
    return air,plant_nodes,e_use


from eagers.config.network import NETWORK_ABBR_NAME_MAP


def update_matrices_step(gen, market, subnet, options, qp, fb, renewable, demand, scale_cost,
        marginal, stor_power, dt, min_power, max_power, step_profile, flt, flc, bat):
    """Update the equalities with the correct demand, and scale fuel and electric costs
     ec is the expected end condition at this time stamp (can be empty)
     stor_power is the expected output/input of any energy storage at this timestep (can be empty)
     min_power and MaxPower define the range of this generator at this timestep
     temperatures is the building and cooling tower water loop temperatures.


     flt    ---     fluid loop temperature
     flc    ---     fluid loop capacitance
     bat    ---     building average temperature
    """
    qp['h'] = [i for i in qp['nom_vectors']['h']]
    qp['f'] = [i for i in qp['nom_vectors']['f']]
    qp['b'] = [i for i in qp['nom_vectors']['b']]
    qp['b_eq'] = [i for i in qp['nom_vectors']['b_eq']]
    qp['lb'] = [i for i in qp['nom_vectors']['lb']]
    qp['ub'] = [i for i in qp['nom_vectors']['ub']]
    qp['const_cost'] = [i for i in qp['nom_vectors']['const_cost']]

    n_g = len(gen)

    # Determine which components are enabled.
    qp['organize']['enabled'] = [True for i in range(n_g)]
    for i in range(n_g):
        if 'enabled' in gen[i] and not gen[i]['enabled']:
            qp['organize']['enabled'][i] = False
    qp['renewable'] = [[] for i in range(len(gen))]
    if len(renewable)>0:
        gen_names = [k['name'] for k in gen]
        for k in renewable:
            qp['renewable'][gen_names.index(k)] = renewable[k]
    

    # Update the demands and self-discharge losses.
    update_mat_demand_step(gen, qp, subnet, options, demand, renewable, stor_power)

    # Building inequalities and equality.
    if fb != None:
        update_mat_building_step(qp, fb, bat, dt)   

    # Fluid loop equality.
    update_fluid_loop_step(flt, flc, qp, dt)

    # Update upper and lower bounds based on ramping constraint (if
    # applicable).
    
    update_gen_limit(gen, qp, min_power, max_power, market)

    # Update storage bounds.
    update_storage_bounds(gen, subnet, qp, stor_power, step_profile, dt)

    # Adding cost for Line Transfer Penalties to differentiate from spillway
    # flow.
    update_mat_hydro_step(gen, qp, subnet)

    # Update costs.
    update_mat_cost_step(gen, qp, step_profile, stor_power, scale_cost, marginal, dt)

    # Update spinning reserve.
    update_spin_reserve_step(qp, options, demand, dt)

    # update market
    update_mat_market_step(qp,options,dt)

    qp['updated_b_eq'] = [i for i in qp['b_eq']]
    qp['updated_lb'] = [i for i in qp['lb']]


def update_mat_demand_step(gen, qp, subnet, options, demand, renewable, stor_power):
    """Update demands and storage self discharge."""
    qp['network'] = {}
    gen_names = [k['name'] for k in gen]
    for net in subnet['network_names']:
        abbrev = subnet[net]['abbreviation']
        qp['network'][net] = {}
        qp['network'][net]['nodes'] = len(subnet[net]['nodes'])
        if net == 'hydro':
            # Don't do a water balance, since it depends on multiple time
            # steps. Any extra outflow at this time step is subtracted from
            # expected outflow at next step (same SOC and flows up river).
            pass
        else:
            dem = demand[net]
            for i in range(len(subnet[net]['nodes'])):
                equip = subnet[net]['equipment'][i]
                r_eq = qp['organize']['balance'][net][i]
                #TODO change to node names rather than index
                for nn in subnet[net]['nodes'][i]:
                    if nn in dem:
                        qp['b_eq'][r_eq] = dem[nn]
                for eq_name in equip:
                    k = gen_names.index(eq_name)
                    if (net == 'electrical' or net == 'direct_current') and eq_name in renewable:
                        # Subtract renewable generation. Put renewable generation into energy balance at correct node.
                        qp['b_eq'][r_eq] -= renewable[eq_name]
                    if not stor_power is None and not stor_power[k] is None:
                        if gen[k]['type'] in ['ElectricStorage', 'ThermalStorage'] and abbrev in gen[k]['output']:
                            alpha = -(1/gen[k]['stor']['charge_eff'] - gen[k]['stor']['disch_eff'])
                            if stor_power[k]>0:
                                qp['b_eq'][r_eq] -= stor_power[k]
                                qp['b'][qp['organize']['inequalities'][k][0]] = - alpha * stor_power[k] / gen[k]['stor']['disch_eff']
                            else:
                                qp['b_eq'][r_eq] -= stor_power[k] * gen[k]['stor']['charge_eff'] * gen[k]['stor']['disch_eff']
                                qp['b'][qp['organize']['inequalities'][k][0]] = - alpha * stor_power[k] * gen[k]['stor']['charge_eff']
                        elif abbrev == 'e' and gen[k]['type'] == 'HydroStorage':
                            qp['b_eq'][r_eq] -= stor_power[k]
            if net == 'electrical' and options['spin_reserve']:
                # -shortfall + SRancillary - SR generators - SR storage <= -SR target
                qp['b'][qp['organize']['spin_reserve'][0]] = -options['spin_reserve_perc']/100 * sum([dem[k] for k in dem])


def update_mat_cost_step(gen, qp, step_profile, stor_power, scale_cost, marginal, dt):    
    n_g = len(gen)
    for i in range(n_g):
        if not qp['indices'][i][0] is None:
            states = qp['indices'][i][0] #states of this generator
            if gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator','TradePoint']:
                for s in states:
                    qp['f'][s] *= scale_cost[i][0] * dt
                    qp['h'][s] *= scale_cost[i][0] * dt
            elif gen[i]['type']  in ['Utility','Market']:
                if gen[i]['source'] == 'electricity':
                    qp['f'][states[0]] *= scale_cost[i][0] * dt
                    if len(states)>1 and (gen[i]['type'] == 'Market' or gen[i]['sellback_rate'] == -1):
                        # Sellback is a fixed percent of purchase costs (percent
                        # set when building matrices).
                        qp['f'][states[1]] *= scale_cost[i][0] * dt
                    else:
                        pass
                        # Constant sellback rate was taken care of when building
                        # the matrices.
            elif gen[i]['type'] == 'HydroStorage':
                if stor_power !=None:
                    qp['f'][states[0]] = marginal['e'] #expected generation moved to RHS, so that cost penalty can refer to deviations from that power
                    qp['h'][0] = 2*qp['f'][states[0]]/(.25*stor_power[i])  #factor of 2 because its solving C = 0.5*x'*H*x + f'*x
            elif 'stor' in gen[i]:
                stor_cat = list(gen[i]['output'].keys())
                # Penalize deviations from the expected storage output.
                a = 1  # Sets severity of quadratic penalty.
                # Scale the penalty by a) larger of 5% capcity and 2x
                # the expected change in stored capacity, b) lesser of
                # a) and remaining space in storage.
                max_charge = 0.05 * gen[i]['stor']['usable_size'] / dt
                if stor_power!=None:
                    max_charge = max([max_charge, 2 * abs(stor_power[i])])
                    us =gen[i]['stor']['usable_size']
                    fp = step_profile[i]
                    if us > fp:
                        max_charge = min([(us- fp) / dt, max_charge])
                    if fp > 0:
                        max_charge = min([0.5 * fp/ dt, max_charge])
                qp['f'][states[0]] = marginal[stor_cat[0]]
                # Factor of 2 because it's solving C = 0.5*x'*h*x + f'*x
                qp['h'][states[0]] = a * 2 * qp['f'][states[0]] / max_charge
                if len(states) > 1:
                    # Small penalty on charging state so that if there
                    # are no other generators it keeps the charging
                    # constraint as an equality.
                    qp['f'][states[1]] = 1e-6
    for i in range(len(qp['const_cost'])):
        if len(scale_cost[i])>0:
            qp['const_cost'][i] *= scale_cost[i][0] * dt


def update_spin_reserve_step(qp, options, demand, dt):
    n_g = len(qp['organize']['dispatchable'])
    if options['spin_reserve']:
        sr_short = qp['organize']['spin_reserve_states'][n_g] #cumulative spinning reserve shortfall 
        # sr_ancillary = qp['organize']['spin_reserve_states'][n_g+1] #Ancillary spinning reserve value (negative cost) 
        if all(demand['electrical'][k] == 0 for k in demand['electrical']):
            spin_cost = 0.02
            print('need building demand for spin reserve cost estimate in update_matrices_step')
        elif options['spin_reserve_perc'] > 5:
            # More than 5% spinning reserve.
            # -shortfall + sr_ancillary - sr_generators - sr_storage <= -sr_target
            spin_cost = 2 * dt / (options['spin_reserve_perc']/(100*sum([demand['electrical'][k] for k in demand['electrical']])))
        else:
            spin_cost = 2*dt/(0.05*sum(demand['e'])) #-shortfall + SRancillary - SR generators - SR storage <= -SR target
        qp['h'][sr_short] = spin_cost #effectively $2 per kWh at point where shortfall = spin reserve percent*demand or $2 at 5%
        qp['f'][sr_short] = 0.05*dt  # $0.05 per kWh.


def update_mat_building_step(qp, fb, temperatures, dt):
    # Within the optimization 5 states are used for a building:
    # #1 represents the air zone temperature setpoint
    # #2 represents the heating required to achieve that zone temperature setpoint
    # #3 represents the cooling required to achieve that zone temperature setpoint
    # #4 represents the temperature in excess of a comfort range (this is penalized)
    # #5 represents the temperature below a comfort range (this is penalized)

    # Building load appears in node equality equation (electric/heating/cooling)
    # 8 inequalities per building:
        #heating: 
            #heating inequality 1 additional heating at lower temp# H>=  UAl*(Th_barl - Ti) + Cap*(Ti - T(i-1))/dt where dt is in seconds
            #heating inequality 2 additional heating at higher temp# H>=  UAi*(Ti-Th_bari) + Cap*(Ti - T(i-1))/dt where dt is in seconds
            #heating inequality 3 minimum heating #  H>=Hmin (done with lower bound)
        #cooling: 
            #Cooling inequality 1 additional cooling at lower temp# C>= UAl*(Tc_barl-Ti) + Cap*(T(i-1)-Ti)/dt where dt is in seconds
            #Cooling inequality 2 additional cooling at higher temp# C>= UAi*(Ti - Tc_bari) + Cap*(T(i-1)-Ti)/dt where dt is in seconds
            #Cooling inequality 3 minimum cooling #  C>=Cmin (done with lower bound)
        #Upper bound inequality
        #Lower bound inequality
        #Positive deviation inequality
        #Negative deviation inequality
    n_b = len(temperatures)
    n_g = len(qp['organize']['dispatchable'])
    n_l = len(qp['organize']['transmission'])
    for i in range(n_b):
        states = qp['indices'][n_g+n_l+i][0]
 
        #electrical energy balance
        r_eq = qp['organize']['building']['electrical_req'][i]
        qp['a_eq'][r_eq][states[1]]= -fb['H2E'][i] # Electricity = E0 + H2E*Heating + C2E*Cooling
        qp['a_eq'][r_eq][states[2]] = -fb['C2E'][i] # Electricity = E0 + H2E*Heating + C2E*Cooling
        qp['b_eq'][r_eq] += fb['E0'][i] #Equipment and nominal Fan Power

        r = qp['organize']['building']['r'][i]
        #heating inequalities
        qp['lb'][states[1]] = fb['h_min'][i]        
        qp['a'][r][states[0]] = -fb['ua_h'][i][0] + fb['Cap'][i]/(3600*dt)
        qp['a'][r+1][states[0]] = fb['ua_h'][i][1] + fb['Cap'][i]/(3600*dt)
        qp['a'][r][states[1]] = -1
        qp['a'][r+1][states[1]] = -1
        qp['b'][r] = -fb['ua_h'][i][0]*fb['th_bar'][i][0] + fb['Cap'][i]/(3600*dt)*temperatures[i]
        qp['b'][r+1] = fb['ua_h'][i][1]*fb['th_bar'][i][1] + fb['Cap'][i]/(3600*dt)*temperatures[i]
        
        #cooling inequalities
        qp['lb'][states[2]] = fb['c_min'][i]
        qp['b'][r+2] = -fb['ua_c'][i][0]*fb['tc_bar'][i][0] - fb['Cap'][i]/(3600*dt)*temperatures[i]
        qp['a'][r+2][states[0]] = -fb['ua_c'][i][0] - fb['Cap'][i]/(3600*dt)
        qp['b'][r+3] = fb['ua_c'][i][1]*fb['tc_bar'][i][1] - fb['Cap'][i]/(3600*dt)*temperatures[i]
        qp['a'][r+3][states[0]] = fb['ua_c'][i][1] - fb['Cap'][i]/(3600*dt)
        qp['a'][r+2][states[2]] = -1
        qp['a'][r+3][states[2]] = -1
        #penalty states (excess and under temperature)
        qp['b'][r+4] = fb['Tmax'][i] #upper buffer inequality, the longer the time step the larger the penaly cost on the state is.
        qp['b'][r+5] = -fb['Tmin'][i] #lower buffer inequality
        #positive and negative deviation inequality
        qp['b'][r+6] = fb['T_avg'][i] #T deviation >= (Ti - Tnominal)
        qp['b'][r+7] = -fb['T_avg'][i] #T deviation >= (Tnominal- Ti)
        
        #Cost penalty for exceeding temperature bounds
        qp['h'][states[3]] = fb['Discomfort'][i]
        qp['h'][states[4]] = fb['Discomfort'][i]
        qp['f'][states[3]] = fb['Discomfort'][i]
        qp['f'][states[4]] = fb['Discomfort'][i]

        #Cost penalty for deviation if deadband in nominal temperature window
        if fb['deadband'][i]:
            qp['f'][states[5]] = fb['Discomfort'][i]
        else:
            qp['f'][states[5]] = 0


def update_fluid_loop_step(flt, flc, qp, dt):
    '''
    flt     ---     fluid loop temperatures
    flc     ---     fluid loop capacitance
    '''

    n_b = len(qp['organize']['building']['r'])
    n_g = len(qp['organize']['dispatchable'])
    n_l = len(qp['organize']['transmission'])
    n_fl = len(flt)
    for i in range(n_fl):
        #1st order temperature model: 0 = -T(k) + T(k-1) + dt/Capacitance*(energy balance)
        state = qp['indices'][n_g+n_l+n_b+i][0]
        r_eq = qp['organize']['balance']['cooling_water'][i]
        # Water capacity in kg and thermal capacitance in kJ/kg*K to get kJ/K.
         #energy balance for chillers & cooling tower fans already put in this row of Aeq
        qp['a_eq'][r_eq][state[0]] = -flc[i]/(3600*dt) #-T(k)
        qp['b_eq'][r_eq] = -flt[i]*flc[i]/(3600*dt) #-T(k-1)


def update_gen_limit(gen, qp, min_power, max_power, market):
    for i in range(len(gen)):
        if qp['organize']['dispatchable'][i]:
        # if gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator'] and not qp['indices'][i][0] is None:
            states = qp['indices'][i][0]
            lb = sum([qp['lb'][s] for s in states])
            if min_power[i] > lb:
                # Raise the lower bounds.
                # qp['organize']['dispatchable'][i] = False  # Can't shut off.
                power_add = min_power[i] - lb
                for f in states:
                    # Starting with lb of 1st state in generator, then moving on.
                    gap = qp['ub'][f] - qp['lb'][f]
                    if gap > 0:
                        gap = min([gap, power_add])
                        qp['lb'][f] += gap
                        power_add -= gap
            if max_power[i] < gen[i]['ub']:
                # Lower the upper bounds.
                power_sub = gen[i]['ub'] - max_power[i]
                for f in range(len(states)-1, -1, -1):
                    # Starting with lb of 1st state in generator, then movingon.
                    if qp['ub'][states[f]] > 0:
                        gap = min(qp['ub'][states[f]], power_sub)
                        qp['ub'][states[f]] -= gap
                        power_sub -= gap
        elif gen[i]['type'] in ['HydroStorage']:
            states = qp['indices'][i][0]
            qp['lb'][states[0]] = min_power[i]
            qp['ub'][states[0]] = max_power[i]
        elif gen[i]['type'] in ['Utility','Tradepoint']:
            pass #already set, shouldn't need to update
        elif gen[i]['type'] in ['Market']:
            states = qp['indices'][i][0]
            #Set the upper and lower bounds to the awarded bid capacity to
            #avoid deviating from commitments
            if market['award']['time'] !=None:
                qp['lb'][states[0]] = 0 #initial market is 0
                qp['ub'][states[0]] = 0 
            else:
                qp['lb'][states[0]] = market['award']['capacity'][len(market['award']['capacity'])+market['award_time'][i]] #gets the award for the current time  
                qp['ub'][states[0]] = market['award']['capacity'][len(market['award']['capacity'])+market['award_time'][i]]


def update_storage_bounds(gen, subnet, qp, stor_power, step_profile, dt):
    n_g = len(gen)
    for i in range(n_g):
        states = qp['indices'][i][0]
        if gen[i]['type'] in ['ElectricStorage', 'ThermalStorage', 'HydroStorage']:
            # Update storage output range to account for what is already scheduled.
            f = list(gen[i]['output'].keys())
            net = NETWORK_ABBR_NAME_MAP[f[0]]
            n = 0
            for equip in subnet[net]['equipment']:
                if gen[i]['name'] in equip:
                    break 
                else:
                    n +=1
            r_eq = qp['organize']['balance'][net][n]#balance at correct node
            s = states[0]
            if stor_power == None: #only when solving automatic initial conditions
                qp['lb'][s] = -gen[i]['stor']['peak_charge']
                qp['ub'][s] = gen[i]['stor']['peak_disch']
                qp['f'][s] = 5*max(qp['f'])
            else:
                charging_space = max([0,(gen[i]['stor']['usable_size']-step_profile[i])*qp['a_eq'][r_eq][s]/dt]) #you can't charge more than you have space for
                qp['lb'][s] = max([qp['lb'][s]-stor_power[i], -charging_space])
                qp['ub'][s] = min([qp['ub'][s]-stor_power[i], step_profile[i]/dt])
            if not qp['organize']['spin_row'][i] is None: #update spinning reserve (max additional power from storage
                qp['b_eq'][qp['organize']['spin_row'][i]] = qp['ub'][s]
            if gen[i]['type'] == 'HydroStorage':
                if stor_power != None:
                    qp['lb'][s] -= stor_power[i]
                    qp['ub'][s] -= stor_power[i]
                    #update the equality with NewPower*conversion + spill - Outflow = nominal PowerGen Flow
                    # n = gen[i]['subnet_node']['hydro']
                    # h = qp['organize']['balance']['hydro'].index(n)
                    qp['b_eq'][qp['organize']['equalities'][i]] = -step_profile[i]*gen[i]['stor']['power_to_flow']
        if not states is None:
            for s in states:
                qp['lb'][s] = min([qp['lb'][s],qp['ub'][s]]) #fix rounding errors


def update_mat_hydro_step(gen, qp, subnet):
    """Also want to penalize spill flow to conserve water for later, spill flow
    is just to meet instream requirments when the power gen needs to be low.
    """

    if 'hydro' in subnet['network_names'] and len(subnet['electrical']['line_names']) > 0:
        n_g = len(gen)
        for k in range(0,len(subnet['electrical']['line_names'])):
            line = subnet['electrical']['line_number'][k]
            states = qp['indices'][n_g + line][0]
            if len(states)>1: #bi-directional transfer with penalty states
                # Adding .1 cent/kWhr cost to every Electric line penalty to
                # differentiate from spill flow.
                qp['f'][states[1]] = 0.001
                qp['f'][states[2]] = 0.001
        
        for i in range(n_g):
            if gen[i]['type'] == 'hydro storage' and 's' in gen[i]:
                states = qp['indices'][i][0]
                qp['f'][states[1]] = 0.0001/gen[i]['stor']['power_to_flow']

def update_mat_market_step(qp,options,dt):
    pass
    # n_g = len(qp['organize']['dispatchable'])
    # if options['market']==True:
    #     market_states = qp['organize']['market']


def fan_calc(fan,air,avail):
    air_density = 1.204  #kg/m^3
    air_out={}
    flow = min(air['m_dot']/air_density,fan['flow_rate']*avail)
    if flow>0:
        if fan['type']=='ConstantVolume' or fan['type']=='OnOff':
            PLF = 1 #should be determined by partial loading of other components in HVAC loop, see page 1040
            RTF = flow/fan['flow_rate']/PLF
            Q_tot = RTF*fan['flow_rate']*fan['pressure_rise']/fan['fan_efficiency'] # m^3/s * Pa = N*m/s = W
        elif fan['type']=='VariableVolume':
            f_flow = flow/fan['flow_rate']
            f_pl = fan['a1'] + fan['a2']*f_flow + fan['a3']*f_flow**2 + fan['a4']*f_flow**3 + fan['a5']*f_flow**4
            Q_tot = f_pl*fan['flow_rate']*fan['pressure_rise']/fan['fan_efficiency'] # m^3/s * Pa = N*m/s = W
        elif fan['type']=='SystemModel':  #designed to replace other types
            pass
        Q_shaft = fan['motor_efficiency']*Q_tot
        Q_to_air = Q_shaft + (Q_tot-Q_shaft)*fan['motor_frac'] #W
        cp = (1006 + 1860*air['w'])*air_density
        air_out['Tdb'] = air['Tdb'] + Q_to_air/(cp*flow)
        air_out['w'] = air['w']+0
        air_out['h'] = air['h'] + Q_to_air/(flow*air_density) #all h in J/kg
        air_out['flow'] = flow+0
        air_out['m_dot'] = flow*air_density
    else:
        air_out = {}
        for k in air:
            air_out[k] = air[k]+0
        Q_tot = 0
        Q_to_air = 0
    return air_out,Q_tot,Q_to_air


from eagers.config.dataset import DATASET_TYPES
from eagers.class_definition.data_set import NodeDataTimeSeries, WeatherTimeSeries

def read_test_data(data):
    fn_args = tuple(data.pop(f'{x}_filename') for x in DATASET_TYPES) # nodedata_filename and weather_filename
    test_data = dict(nodedata=NodeDataTimeSeries(fn_args[0]), weather=WeatherTimeSeries(fn_args[1]))
    test_data['nodedata_network_info'] = data['nodedata_network_info']
    test_data['hist_prof'] = []
    test_data['_nodedata_filename'] = test_data['nodedata'].filepath.name
    test_data['_weather_filename'] = test_data['weather'].filepath.name
    return test_data


from eagers.building.basic.window_solar_transmittance import window_solar_transmittance
from eagers.building.basic.solar_calc import solar_calc
from eagers.building.basic.area_weighted_temperature import area_weighted_temperature
from eagers.building.process.load_sched import load_sched
from eagers.building.process.zone_loads import zone_loads

import numpy as np


def load_actual_building_data(buildings,observer,weather,forecast,date,res):
    n_b = len(buildings)
    if n_b>0:
        actual_data = {}
        for t in range(len(date)):
            weather_now = {}
            for i in weather:
                weather_now[i] = weather[i][t]
            internal_gain_scalar = building_variability(buildings)
            
            actual_data['loads'] = [{} for i in range(n_b)]
            actual_data['gains'] = [{} for i in range(n_b)]
            actual_data['S'] = [[] for i in range(n_b)]
            actual_data['Tz_nominal'] = [[] for i in range(n_b)]
            actual_data['T_avg'] = [[] for i in range(n_b)]
            actual_data['E0'] = [forecast['E0'][0]]
            actual_data['H0'] = [forecast['H0'][0]]
            actual_data['C0'] = [forecast['C0'][0]]
            for i, bldg in enumerate(buildings):
                humidity = observer['building_humidity'][i]
                T_zone = observer['building_zone_temp'][i]
                frost = observer['building_frost'][i]
                schedules = load_sched(bldg.schedule, bldg.holidays, date, None, False)#can replace with modified schedules
                loads, gains, _, _, _, _, _, _, _ = zone_loads(bldg, date[0], schedules, weather_now, T_zone, humidity, res, frost)
                actual_gains = [internal_gain_scalar[i][j]*g for j,g in enumerate(loads['internal_gain'])]
                new_loads, new_gains,S = update_real_gains(bldg, weather_now, loads, gains, date[0], actual_gains)
                actual_data['loads'][i] = {}
                for k in new_loads:
                    if isinstance(new_loads[k],dict):
                        actual_data['loads'][i][k] = {}
                        for kk in new_loads[k]:
                            if isinstance(new_loads[k][kk],list):
                                actual_data['loads'][i][k][kk] = [j for j in new_loads[k][kk]]
                            else:
                                actual_data['loads'][i][k][kk] = new_loads[k][kk]+0
                    elif isinstance(new_loads[k],list):
                        actual_data['loads'][i][k] = [j for j in new_loads[k]]
                    else:
                        actual_data['loads'][i][k] = new_loads[k]+0
                actual_data['gains'][i] = {}
                for k in new_gains:
                    if isinstance(new_gains[k],list):
                        actual_data['gains'][i][k] = [j for j in new_gains[k]]
                    else:
                        actual_data['gains'][i][k] = new_gains[k]+0
                actual_data['S'][i] = [[j for j in w] for w in S]
                actual_data['Tz_nominal'][i] = [j for j in forecast['Tz_nominal'][i]]
                actual_data['T_avg'][i] = [area_weighted_temperature(bldg,actual_data['Tz_nominal'][i])]
    else:
        actual_data = None
    return actual_data


def building_variability(buildings):
    ##TODO replace with values or function to add variability to schedule
    scale = [[] for i in range(len(buildings))]
    for i, bldg in enumerate(buildings):
        scale[i] = [1 for z in range(len(bldg.zones['name']))] 
    return scale


def update_real_gains(building,weather,loads,gains,date,new_internal_gain):
    '''
    pulls in actual weather and 'actual' internal gains to scale the original
    lighting, plug load and occupancy gains of each zone. The gains are then
    split to convected and latent for use in the building simulation
    '''
    new_loads = {}
    new_gains = {}
    ## Solar
    _,_, azimuth, zenith = solar_calc(building.location['longitude'][0],building.location['latitude'][0],building.location['timezone'][0],[date])
    vect_2_sun = [float(np.sin(np.radians(zenith[0]))*np.sin(np.radians(azimuth[0]))),
                  float(np.sin(np.radians(zenith[0]))*np.cos(np.radians(azimuth[0]))),
                  float(np.cos(np.radians(zenith[0])))]## normal vector pointed at sun from azimuth (compass heading of sun) and zenith (angle beween sun and normal to surface of earth)
    #solar gain per unit area of exterior        
    if (weather['dir_norm_irr']+weather['dif_horz_irr'])>0:
        ext_gain = []
        for i in range(len(building.surfaces['exterior']['cos_phi'])):
            if building.surfaces['exterior']['boundary'][i]=='ground':
                ext_gain.append(0)
            else:    
                F_s = 0.5*(1+building.surfaces['exterior']['cos_phi'][i])
                F_g = 0.5*(1-building.surfaces['exterior']['cos_phi'][i])
                ext_gain.append(weather['dif_horz_irr']*(building.surfaces['exterior']['solar_absorptance'][i]*F_s) + building.site['ground_reflect']*weather['dif_horz_irr']*(building.surfaces['exterior']['solar_absorptance'][i]*F_g))
                cos_theta = vect_2_sun[0]*building.surfaces['exterior']['normal'][i][0] + vect_2_sun[1]*building.surfaces['exterior']['normal'][i][1] + vect_2_sun[2]*building.surfaces['exterior']['normal'][i][2] #dot product of two vectors
                SF = 1 #shading factor, needs development, pg 211
                if cos_theta>0:
                    ext_gain[i] += weather['dir_norm_irr']*building.surfaces['exterior']['solar_absorptance'][i]*SF*cos_theta #Absorbed direct normal radiation
    else:
        ext_gain = [0 for i in range(len(building.surfaces['exterior']['cos_phi']))]  
    

    ## scale loads to actual internal gains
    nz = len(new_internal_gain)
    scale = []
    for i in range(nz):
        if loads['internal_gain'][i]>0:
            scale.append(new_internal_gain[i]/loads['internal_gain'][i])
        else:
            scale.append(0)
    
    new_loads['latent'] = [loads['latent'][i]*scale[i] for i in range(nz)]
    new_loads['convected'] = [loads['convected'][i]*scale[i] for i in range(nz)]
    new_loads['visible'] = [loads['visible'][i]*scale[i] for i in range(nz)]
    new_loads['radiant'] = [loads['radiant'][i]*scale[i] for i in range(nz)]
    new_loads['lighting_internal'] = [loads['lighting_internal'][i]*scale[i] for i in range(nz)]
    new_loads['plug_load'] = [loads['plug_load'][i]*scale[i] for i in range(nz)]
    new_loads['exterior'] = {}
    new_loads['exterior']['lighting'] = loads['exterior']['lighting']+0
    new_loads['exterior']['equipment'] =  loads['exterior']['equipment']+0
    new_loads['case'] = {}
    new_loads['case']['electric'] = [j for j in loads['case']['electric']]
    new_loads['rack'] = {}
    new_loads['rack']['electric'] = [j for j in loads['rack']['electric']]
    new_loads['multiplier'] = [j for j in loads['multiplier']]

    ## Windows
    visible = [loads['visible'][building.windows['zone_index'][i]]/building.zones['interior_area'][building.windows['zone_index'][i]] for i in range(len(building.windows['area']))]
    long_wave = [loads['radiant'][building.windows['zone_index'][i]]/building.zones['interior_area'][building.windows['zone_index'][i]] for i in range(len(building.windows['area']))]
    window_gain,S = window_solar_transmittance(building.windows,building.surfaces,vect_2_sun,weather['dir_norm_irr'],weather['dif_horz_irr'],visible,long_wave)

    new_gains['zone_sensible'] = [gains['zone_sensible'][i] - loads['convected'][i] + new_loads['convected'][i]  for i in range(nz)]
    new_gains['zone_latent'] = [gains['zone_latent'][i] - loads['latent'][i] + new_loads['latent'][i] for i in range(nz)]
    new_gains['interior_surface'] = [(new_loads['radiant'][building.surfaces['zone_index'][i]] + new_loads['visible'][building.surfaces['zone_index'][i]])*building.surfaces['area'][i]/building.zones['interior_area'][building.surfaces['zone_index'][i]] for i in range(len(building.surfaces['area']))]
    new_gains['exterior_surface'] = [j for j in ext_gain]
    new_gains['hvac_sensible'] = [j for j in gains['hvac_sensible']]
    new_gains['hvac_latent'] = [j for j in gains['hvac_latent']]
    new_gains['window'] = [j for j in window_gain]
    new_gains['ref_rack_water_sensible'] = [j for j in gains['ref_rack_water_sensible']]
    new_gains['ref_rack_water_latent'] = [j for j in gains['ref_rack_water_latent']]
    return new_loads,new_gains,S


''' https://www.nwd.usace.army.mil/CRWM/'''

import urllib.request
import time
from datetime import datetime

from eagers.config.path_spec import USER_DIR_DATA_RETRIEVAL


def query_web(html_list,file_list):
    for i in range(len(html_list)):
        urllib.request.urlretrieve(html_list[i],file_list[i])
        time.sleep(1)

def build_lists(sites,param,units,save_dir,d1,d2):
    #build list of URL's and file names
    # Note that electric data only available older than 6 days
    base = 'https://www.nwd-wc.usace.army.mil/dd/common/web_service/webexec/ecsv?id='
    d_now = datetime.now()
    days = (d_now - d1).days
    if d_now.minute>30:
        tod = 'd' + str(d_now.hour) + 'h' + str(d_now.minute-30) + 'm'
    else:
        tod = 'd' + str(d_now.hour-1) + 'h' + str(d_now.minute+30) + 'm'
    weeks = int(days/7)
    days -= 7*weeks
    lookback = '&headers=true&filename=&timezone=PST&lookback=' + str(weeks) + 'w' + str(days) + tod
    days = (d_now - d2).days
    weeks = int(days/7)
    days -= 7*weeks
    lookforward = '&lookforward=' + '-' + str(weeks) + 'w' + str(days) + tod
    m = d1.month
    if m<10:
        m = '0' + str(m)
    else:
        m = str(m)
    day = d1.day
    if day<10:
        day = '0' + str(day)
    else:
        day = str(day)
    start_date = '&startdate=' + m + '%2F' + day + '%2F' + str(d1.year) + '+07%3A00'
    m = d2.month
    if m<10:
        m = '0' + str(m)
    else:
        m = str(m)
    day = d2.day
    if day<10:
        day = '0' + str(day)
    else:
        day = str(day)
    end_date = '&enddate=' + m + '%2F' + day + '%2F' + str(d2.year) + '+07%3A00'
    html_list = []
    file_list = []    
    for k in range(len(sites['names'])):
        code = sites['abbrev'][k]
        tags = ''
        for j in range(len(param)):
            tags = tags + code + param[j] + units[j]
        url = base + tags + lookback + lookforward + start_date + end_date
        f_name = save_dir + '/' + sites['names'][k]  + '.txt'
        html_list.append(url)
        file_list.append(f_name)
    return html_list, file_list

save_dir = USER_DIR_DATA_RETRIEVAL
sites = {}
sites['names'] = ['bonnevile', 'dalles', 'john_day', 'mcnary', 'priest_rapids', 'wanapum', 'rock_island', 'rocky_reach', 'wells', 'chief_joseph', 'grand_coulee', 'ice_harbor', 'lower_monumental', 'little_goose', 'lower_granite', 'dworshak', 'albeni_falls']
sites['abbrev'] = ['BON', 'TDA', 'JDA', 'MCN', 'PRD', 'WAN', 'RIS', 'RRH', 'WEL', 'CHJ', 'GCL', 'IHR', 'LMN', 'LGS', 'LWG', 'DWR', 'ALF']
start_date = datetime(2019, 10, 1)
end_date = datetime(2020, 9, 30)
param = ['.Flow-Gen.Ave.1Hour.1Hour.CBT-REV','.Flow-Spill.Ave.1Hour.1Hour.CBT-REV','.Flow-Out.Ave.1Hour.1Hour.CBT-REV','.Power.Total.1Hour.1Hour.CBT-RAW']
flow_unit = '%3Aunits%3Dkcfs'
power_unit = '%3Aunits%3DMW'
units = [flow_unit + '%7C', flow_unit + '%7C', flow_unit + '%7C', power_unit]
html_list,file_list = build_lists(sites,param,units,save_dir,start_date,end_date)
query_web(html_list,file_list)#query website

# Additional flow only data from USACE
sites['names'] = ['hells_canyon_outflow']
sites['abbrev'] = ['HCDI']
param = ['.Flow.Inst.15Minutes.0.USGS-REV']
units = [flow_unit]
html_list,file_list = build_lists(sites,param,units,save_dir,start_date,end_date)
query_web(html_list,file_list)#query website

sites['names'] = ['cabinet_gorge_outflow', 'noxon_outflow']
sites['abbrev'] = ['CAB', 'NOX']
param = ['.Flow-Out.Ave.1Hour.1Hour.CBT-RAW','.Flow-In.Ave.1Hour.0.CENWS-COMPUTED-RAW']
units = [flow_unit + '%7C', flow_unit]
html_list,file_list = build_lists(sites,param,units,save_dir,start_date,end_date)
query_web(html_list,file_list)#query website

sites['names'] = ['cabinet_gorge_inflow']
sites['abbrev'] = ['CAB']
param = ['.Flow-In.Ave.1Hour.0.CENWS-COMPUTED-RAW']
units = [flow_unit]
html_list,file_list = build_lists(sites,param,units,save_dir,start_date,end_date)
query_web(html_list,file_list)#query website

# Additional storage data from USACE
sites['names'] = ['cabinet_gorge_storage','noxon_storage']
sites['abbrev'] = ['CAB','NOX']
param = ['.Stor.Inst.1Hour.0.CBT-REV']
units = ['%3Aunits%3Dkaf']
html_list,file_list = build_lists(sites,param,units,save_dir,start_date,end_date)
query_web(html_list,file_list)#query website


'''Idaho Power data'''
'https://idastream.idahopower.com/Data/List/Parameter/Flow/Statistic/LATEST%20Flow/Interval/Latest'
'https://www.idahopower.com/community-recreation/recreation/water-information/stream-flow-data/'

''' BC Hydro'''
'https://wateroffice.ec.gc.ca/google_map/google_map_e.html?map_type=historical&search_type=region&region=PYR'


'''
Plot as validation against energy plus run
'''
import csv
import matplotlib.pyplot as plt 
from multiprocessing import Process

def plot_sim(facility,T_zone,Q_transfer,building,date,filename):
    dr=csv.DictReader(open(filename,'r'))
    e_plus = {}
    for row in dr:
        if len(list(e_plus.keys())) == 0:
            e_plus={k.rstrip():[] for k in row}
        for k, v in row.items():
            if k != 'Date/Time':
                e_plus[k.rstrip()].append(float(v))
    e_plus['Date/Time'] = building.sim_date[1:]
    EP = ['Electricity:Facility [J](Hourly)','Heating:Gas [J](Hourly)','Heating:Electricity [J](Hourly)','WaterSystems:Gas [J](Hourly)','Cooling:Electricity [J](Hourly)','HeatRejection:Electricity [J](Hourly)','Fans:Electricity [J](Hourly)','Pumps:Electricity [J](Hourly)']
    EAG = ['net_elec','heat_gas','heat_elec','water_gas','cool_elec','tower_elec','fan_elec','pumps']
    label_x = ['electric ---total (kw)','heating --- gas (kW)','heating --- electric (kW)','water system --- gas (kW)','cooling (kW)','cooling tower fans (kW)','HVAC fans (kW)','water pumps --- (kW)']

    i1 = 0
    while building.sim_date[i1+1]<date[0]:
        i1+=1
    i2 = i1+0
    while building.sim_date[i2+1]<=date[-1]:
        i2+=1
    model_fit = {}
    for j in range(len(EP)):
        if EP[j] in e_plus and any([e_plus[EP[j]][i]>0 for i in range(i1,i2)]):
            x = [e_plus[EP[j]][i]/3600/1000 for i in range(i1,i2)]
            y = [i/1000 for i in facility[EAG[j]]]
            m_x = sum(x)/len(x)
            m_y = sum(y)/len(y)
            if sum([(x[i]-m_x)**2 for i in range(len(x))]) == 0:
                COD = 0
            else:
                COD = 1 - sum([(y[i] - x[i])**2 for i in range(len(x))])/sum([(x[i]-m_x)**2 for i in range(len(x))]) #Coefficient of determination
            if sum([(y[i]-m_y)**2 for i in range(len(x))]) == 0:
                COD_B = 0
            else: 
                COD_B = 1 - sum([(y[i] - x[i])**2 for i in range(len(x))])/sum([(y[i]-m_y)**2 for i in range(len(x))])
            if (sum([(x[i]-m_x)**2 for i in range(len(x))])*sum([(y[i]-m_y)**2 for i in range(len(x))])) == 0:
                pearson = 0
            else:
                pearson = sum([(x[i]-m_x)*(y[i] - m_y) for i in range(len(x))])**2/(sum([(x[i]-m_x)**2 for i in range(len(x))])*sum([(y[i]-m_y)**2 for i in range(len(x))])) #Pearson product moment correlation coefficient
            model_fit[EAG[j]] = {}
            model_fit[EAG[j]]['Pearson'] = pearson
            model_fit[EAG[j]]['COD'] = max(COD,COD_B)
            model_fit[EAG[j]]['total_energy_percent'] = sum(y)/sum(x)
            p = Process(target=plot_fig, args=(j,date,x,y,label_x[j]))
            p.start()

    p_n = len(EP)+2
    plot_z = [i for i in range(len(building.zones['name']))]
    for p in range(len(plot_z)):
        name = building.zones['name'][plot_z[p]].upper()
        field = 'Cooling:EnergyTransfer:Zone:' + name +' [J](Hourly)'
        if len(field)>63:
            field = field[:63]
        x = [e_plus[field][i]/3600/1000 for i in range(i1,i2)]
        y = [max(0,-i[plot_z[p]]/1000) for i in Q_transfer]
        y_lab = building.zones['name'][plot_z[p]] + ' zone cooling (kW)'
        p = Process(target=plot_fig, args=(p_n+p,date,x,y,y_lab))
        p.start()

    p_n = p_n + len(plot_z)
    for p in range(len(plot_z)):
        name = building.zones['name'][plot_z[p]].upper()
        field = 'Heating:EnergyTransfer:Zone:'+ name + ' [J](Hourly)'
        if len(field)>63:
            field = field[:63]
        x = [e_plus[field][i]/3600/1000 for i in range(i1,i2)]
        y = [max(0,i[plot_z[p]]/1000) for i in Q_transfer]
        y_lab = building.zones['name'][plot_z[p]] + ' zone heating (kW)'
        p = Process(target=plot_fig, args=(p_n+p,date,x,y,y_lab))
        p.start()

    p_n = p_n + len(plot_z)
    for p in range(len(plot_z)):
        name = building.zones['name'][plot_z[p]].upper()
        field = name + ':Zone Air Temperature [C](Hourly)'
        if len(field)>63:
            field = field[:63]
        x = [e_plus[field][i] for i in range(i1,i2)]
        y = [T_zone[i+1][plot_z[p]] for i in range(len(date))]
        y_lab = building.zones['name'][plot_z[p]] + ' zone temperature (C)'
        p = Process(target=plot_fig, args=(p_n+p,date,x,y,y_lab))
        p.start()

def plot_fig(p,d,x,y,y_lab):
    m_x = sum(x)/len(x)
    m_y = sum(y)/len(y)
    if sum([(x[i]-m_x)**2 for i in range(len(x))]) == 0:
        COD_A = 0
    else:
        COD_A = 1 - sum([(y[i] - x[i])**2 for i in range(len(x))])/sum([(x[i]-m_x)**2 for i in range(len(x))]) #Coefficient of determination
    if sum([(y[i]-m_y)**2 for i in range(len(x))]) == 0:
        COD_B = 0
    else: 
        COD_B = 1 - sum([(y[i] - x[i])**2 for i in range(len(x))])/sum([(y[i]-m_y)**2 for i in range(len(x))])
    COD = max(COD_A,COD_B)
    plt.figure(p)
    plt.plot(d,y, label = 'WSU')
    plt.plot(d,x,label = 'EnergyPlus')
    plt.xlabel('Day of year')
    plt.ylabel(y_lab)
    plt.annotate('r^2 value :'+str(COD),xy = (x[0],y[0]),xytext = [.05, .95],xycoords= 'figure fraction')
    plt.annotate('NF value :'+str(sum(y)/sum(x)),xy = (x[0],y[0]),xytext = [.55, .95],xycoords= 'figure fraction')
    plt.ioff()
    plt.draw()
    plt.show()

